;This is the source code for SpeechLib, a machine language library
;for accessing MacinTalk from within Microsoft BASIC

;Kevin J. Willey
 
;15 March 1986
;Version 3.0a

;Provides access to all of MacinTalk
;+ Translation function to convert English to Phonemes
;+ Command to change all parameters back to defaults

;|=========================================================================|
;|    BASIC Syntax                                                         |
;|=========================================================================|
;|    Speech 1[,"Exception Filename"]           calls SpeechOn             |
;|     if Filename='noReader', then only Phonetic input is allowed.        |
;|-------------------------------------------------------------------------|
;|    Speech 2,"Text to Speak"                  calls Reader and MacinTalk |
;|     unless Phonetic input was specified then calls only Macintalk       |
;|-------------------------------------------------------------------------|
;|    Speech 3                                  calls SpeechOff            |
;|-------------------------------------------------------------------------|
;|    Speech 4,RATE     (85 <= RATE  <= 425)    calls SpeechRate           |
;|-------------------------------------------------------------------------|
;|    Speech 5,PITCH    (65 <= PITCH <= 500)    calls SpeechPitch          |
;|-------------------------------------------------------------------------|
;|    Speech 6,MODE      MODE=0:Natural         calls SpeechPitch          |
;|                           =1:Robotic          (but sets MODE)           |
;|-------------------------------------------------------------------------|
;|    Speech 7,SEX       reserved for future implementation                |
;|-------------------------------------------------------------------------|
;|    Speech 8,LANGUAGE  reserved for future implementation                |
;|-------------------------------------------------------------------------|
;|    Speech 9           sets all parameters back to defaults              |
;|-------------------------------------------------------------------------|
;|    Speech 10,"English to Translate",Phoneme$                            |
;|-------------------------------------------------------------------------|
;|    Speech 11          pronounces infostring                             |
;|=========================================================================|


;External References (these are the routines in SpeechAsm.Rel)
;~~~~~~~~~~~~~~~~~~~
XREF	SpeechOn        ;open driver and exception file
XREF	MacinTalk       ;pronounce phonemes
XREF	Reader          ;translate English to phonemes
XREF	SpeechOff       ;close the driver and clean up memory
XREF	SpeechRate      ;set the speaking rate
XREF	SpeechPitch     ;set the baseline pitch and/or mode



;TRAP calls
;~~~~~~~~~~
.TRAP	_NewHandle	$A122   ;INSIDE MACINTOSH (promo ed.) lists $A022
.TRAP	_DisposHandle	$A023   
.TRAP	_BlockMove	$A02E
.TRAP	_CmpString	$A03C
.TRAP	_PtrToHand	$A9E3
.TRAP	_GetHandleSize	$A025


;jump table for BASIC routines {pointed to by a5}
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
GetNextLibArg	EQU $2A	;retrieves next argument from input list
IntegerArg	EQU $32	;converts argument into 16 bit signed integer
BasicError	EQU $42	;generates error code indicated by d2
AllocString	EQU $72	;creates space for a string
AllocTempDesc	EQU $7A	;generates a temporary string desc for a string
FreeTempDesc	EQU $82	;frees temporary string descriptors from memory
LoadStringDesc	EQU $8A	;loads contents of string descriptor in registers
AssignString	EQU $92	;assigns a temporary string to a string variable



;SpeechLib Routines

;##########################################################################
Speech:

;----- Parse the Function Number ------------------------------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~~	
;Get the input argument from BASIC and branch to the correct code

	JSR	GetIntegerArg	;Comparison Table to parse chosen function
	CMP.B	#1,d3
	BEQ	InitSpeech
	CMP.B	#2,d3
	BEQ	Talk
	CMP.B	#3,d3
	BEQ	QuitSpeech
	CMP.B	#4,d3
	BEQ	SetRate
	CMP.B	#5,d3
	BEQ	SetPitch
	CMP.B	#6,d3
	BEQ	SetMode
	CMP.B	#7,d3
	BEQ	SetSex
	CMP.B	#8,d3
	BEQ	SetLanguage
	CMP.B	#9,d3
	BEQ	SetDefault
	CMP.B	#10,d3
	BEQ	Translate
	CMP.B	#11,d3
	BEQ	GiveInfo
	MOVE.L	#9,d2           ;"SUBSCRIPT OUT-OF-RANGE" error preloaded
	BRA	ErrorReport     ;report error
;##########################################################################


;##########################################################################
InitSpeech:

;----- Load the Speech Driver and Speech Globals --------------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	JSR	GetNextLibArg(a5)   ;get argument from BASIC
	CMP.B	#0,d0               ;check if end of argument list
	BEQ	NoRules             ;if no argument then use default rules
	BSR	ParseStringArg      ;move string descriptor into regs.
	BSR	GenerateFilename    ;convert BASIC string to STR255
	BRA	OpenDriver

NoRules:
;  use default pronunciation rules
	LEA	ExceptionFile,a1
	MOVE.L	#0,(a1)
		
OpenDriver:

;FUNCTION SpeechOn (ExceptionFile:STR255; VAR theSpeech:SpeechHandle):
;                   SpeechErr;
	
	CLR.W	-(SP)          ;clear space for result code
	PEA	ExceptionFile  ;load Exception File
	PEA	theSpeech      ;load speech globals
	JSR	SpeechOn       ;jump to driver open routine
	MOVE.W	(SP)+,d0       ;check resulting error code
	MOVE.L	#7,d2          ;d2 pre-loaded with "OUT OF MEMORY" error
	CMP.W	#0,d0          ;if error code=0 then open worked
	BNE	ErrorReport    ;"OUT OF HEAP SPACE" error pre-loaded in d2
	BRA	CloseItUp      ;Return to BASIC
;##########################################################################


;##########################################################################
Talk:

;----- Check for English or Phonetic Input --------------------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;PhonemeFlag was either set or cleared depending on the exception file used

	LEA	PhonemeFlag,a1
	TST.B	(a1)		;if Phoneme flag is clear then convert
	BEQ	English		;English to Phonemes


Phonemes:
	BSR	GenPHandle	;Create a handle and fill it with phonemes
	BSR	Pronounce	;Pass the phonemes to MacinTalk for speech	
	BRA	CleanMem	;Clean Up


English:
	BSR	GenEHandle	;Create an Empty Handle	
	BSR	Read		;Fill with translated English phonemes
	BSR	Pronounce	;Pass the phonemes to MacinTalk for speech
	BRA	CleanMem	;Clean Up


;*********************** Subroutines for Talking ***************************
;***************************************************************************

;----- Generate a handle to the input Phonetic string ---------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
GenPHandle:

;FUNCTION PtrToHandle (srcPtr:Ptr; VAR dstHndl:Handle; size:LongInt):OSErr;
;on Input  : a0=srcPtr, a1=dstHndl, d0=size
;on Output : a1=dstHndl, d0=result code

	BSR	GetStringArg	;get input string
	MOVE.L	a3,a0		;set up registers for _PtrToHand conversion
	MOVE.L	d2,d0
Info:	MOVE.L	#7,d2		;pre-load an "OUT OF MEMORY" error
	_PtrToHand		;create a handle to the Phonemes
	BMI	ErrorReport	;report error
	LEA	phHandle,a1	;copy handle into phHandle
	MOVE.L	a0,(a1)
	RTS

;----- Generate a handle for the translated English text ------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
GenEHandle:

;FUNCTION NewHandle (logicalSize:Size):Handle
;on Input  : d0=logicalSize
;on Output : a0=Handle, d0=result code
	
	MOVEQ	#0,d0         ;set up empty handle
	_NewHandle            ;allocate handle for reader to fill 
                              ;with phonemes (output:a0 = the handle)
	LEA	phHandle,a1   
	MOVE.L	a0,(a1)       ;save handle in phHandle
	RTS

;----- Translate the English text ------------------------------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Read:

;FUNCTION Reader (theSpeech:SpeechHandle; EnglishInput:Ptr
; 		  InputLength:LongInt; PhoneticOutput:Handle):SpeechErr;

	CLR.W	-(SP)           ;clear space for result
	MOVE.L	theSpeech,-(SP) ;load speech globals
	BSR	GetStringArg    ;Get text to be spoken from BASIC
	MOVE.L	a3,-(SP)        ;load Ptr to string
	MOVE.L	d2,-(SP)        ;load String length
	MOVE.L	phHandle,-(SP)  ;we just allocated this handle for phonemes
	JSR	Reader          ;do translation of English to phonemes
	MOVE.W	(SP)+,d0        ;get resulting error code
	MOVE.L	#7,d2           ;"OUT OF HEAP SPACE" error pre-loaded in d2
	CMP.W	#0,d0           ;if error code=0 then Reader worked
	BNE	ErrorReport     ;report error
	RTS


;----- Pronounce a string of Phonemes -------------------------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Pronounce:

;now feed the phonemes to Macintalk
;FUNCTION MacinTalk (theSpeech:SpeechHandle; Phonemes:Handle):SpeechErr
	
	CLR.W	-(SP)           ;space for result code
	MOVE.L	theSpeech,-(SP) ;load speech globals
	MOVE.L	phHandle,-(SP)  ;load handle to phonemes
	JSR	MacinTalk       ;say it
	MOVE.W	(SP)+,d0        ;get result code
	MOVE.L	#3,d2           ;"DEVICE UNAVAILABLE" error preloaded in d2
	CMP.B	#0,d0           ;if error code=0 then MacinTalk worked
	BNE	ErrorReport     ;report error
	RTS


;----- Dispose of memory blocks -------------------------------------------
;      ~~~~~~~~~~~~~~~~~~~~~~~~
CleanMem:

;PROCEDURE DisposHandle (h:Handle)
;on Input  : a0=handle
	MOVE.L	phHandle,a0     ;_DisposHandle expects input in a0
	_DisposHandle           ;deallocate handle to free up memory
	BRA CloseItUp           ;return to BASIC
;##########################################################################


;##########################################################################
QuitSpeech:

;------ Clean up memory and leave -----------------------------------------
;       ~~~~~~~~~~~~~~~~~~~~~~~~~
;PROCEDURE SpeechOff (theSpeech:SpeechHandle)

	MOVE.L	theSpeech,-(SP) ;handle to speech globals
	JSR	SpeechOff       ;close it up
	MOVE.L	#0,d0           ;exit with 0 in d0 makes code purgeable
	                        ;          1 in d0 makes code resident
	RTS
;##########################################################################


;##########################################################################
SetRate:

;------ Set the speaking rate ---------------------------------------------
;       ~~~~~~~~~~~~~~~~~~~~~
;PROCEDURE SpeechRate (theSpeech:SpeechHandle; theRate:INTEGER);

	BSR	GetIntegerArg   ;Get user input RATE {returned in d3}
	MOVE.L	#9,d2           ;"SUBSCRIPT OUT-OF-RANGE" error pre-loaded
	CMP.W	#85,d3          ;check lower bound
	BLT	ErrorReport     ;report error
	CMP.W	#425,d3         ;check upper bound
	BGT	ErrorReport     ;report error
	MOVE.L	theSpeech,-(SP) ;load speech globals
	MOVE.W	d3,-(SP)        ;load RATE - {RATE is contained in d3...
	JSR	SpeechRate      ;set the RATE	
	BRA	CloseItUp       ;return to BASIC
;##########################################################################

;##########################################################################
SetPitch:

;------ Set the baseline pitch --------------------------------------------
;       ~~~~~~~~~~~~~~~~~~~~~~
;PROCEDURE SpeechPitch (theSpeech:SpeechHandle; thePitch:INTEGER;
;			theMode:FOMode);    
; 	- theMode is NoChange in order to only affect PITCH

	BSR	GetIntegerArg   ;Get user input PITCH {returned in d3}
	MOVE.L	#9,d2           ;"SUBSCRIPT OUT-OF-RANGE" error preloaded
	CMP.W	#65,d3          ;check lower bound
	BLT	ErrorReport     ;report error
	CMP.W	#500,d3         ;check upper bound
	BGT	ErrorReport     ;report error
	MOVE.L	theSpeech,-(SP) ;speech globals
	MOVE.W	d3,-(SP)        ;Pitch
	MOVE.B	#2,-(SP)        ;Specify 'NoChange' for the MODE
	JSR	SpeechPitch     ;change SpeechPitch
	BRA	CloseItUp       ;return to BASIC
;##########################################################################


;##########################################################################
SetMode:

;------ Set the speech mode -----------------------------------------------
;       ~~~~~~~~~~~~~~~~~~~
;PROCEDURE SpeechPitch (theSpeech:SpeechHandle; thePitch:INTEGER;
;			theMode:FOMode);
;  set PITCH out of range and PITCH will default to previous value
	
	BSR	GetIntegerArg   ;Get user selected MODE {returned in d3}
	MOVE.L	#9,d2           ;"SUBSCRIPT OUT-OF-RANGE" error preloaded
	CMP.B	#1,d3           ;check upper bound
	BGT	ErrorReport     ;report error
	MOVE.L	theSpeech,-(SP) ;speech globals
	MOVE.W	#0,-(SP)        ;PITCH out of range so no change to SpeechPitch	
	MOVE.B	d3,-(SP)        ;set MODE - MODE is the ordinality of the 
	                        ;elements of the Pascal data type F0Mode.
	JSR	SpeechPitch     ;set the MODE
	BRA	CloseItUp       ;return to BASIC	
;##########################################################################

	
;##########################################################################
SetSex:

;------ Set the speech sex -------------------------------------------------
;       ~~~~~~~~~~~~~~~~~~
;PROCEDURE SpeechSex (theSpeech:SpeechHandle; theSex:SEX);
	
	BRA	CloseItUp       ;return to BASIC	
;##########################################################################


;##########################################################################
SetLanguage:

;------ Set the speech language --------------------------------------------
;       ~~~~~~~~~~~~~~~~~~~~~~~
;PROCEDURE SpeechLanguage (theSpeech:SpeechHandle; ?:?);
	
	BRA	CloseItUp       ;return to BASIC	
;##########################################################################


;##########################################################################
SetDefault:
	MOVE.L	theSpeech,-(SP)	;load speech globals
	MOVE.W	#150,-(SP)	;load RATE
	JSR	SpeechRate	;set the RATE	
	MOVE.L	theSpeech,-(SP) ;load speech globals
	MOVE.W	#110,-(SP)	;load PITCH
	MOVE.B	#0,-(SP)	;load MODE
	JSR	SpeechPitch	;set the PITCH and the MODE
	BRA	CloseItUp
;##########################################################################


;##########################################################################
Translate:

;------ Input English and Output Phonetic code -----------------------------
;       ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	LEA	PhonemeFlag,a1	;if Phoneme flag is clear then convert
	TST.B	(a1)		;English to Phonemes, else leave and do
	BNE	CloseItUp	;nothing
				
	BSR	GenEHandle		;create empty handle for phonemes
	BSR	Read			;convert English to Phonemes
	MOVE.L	phHandle,a0		;set up regs. for _GetHandleSize
	_GetHandleSize			;get length of Phoneme string
	MOVE.W	d0,d2			;use same length for AllocString
	JSR	AllocString(a5)		;create an empty string
	MOVE.L	a2,a1			;set up regs. for _BlockMove
	MOVE.L	(a0),a0			;de-reference Phoneme handle
	_BlockMove			;copy Phonemes to empty string
	JSR	AllocTempDesc(a5)	;create a string desc. for string
	MOVE.L	a3,a1			;save pointer to new string desc.
	JSR	GetNextLibArg(a5)	;get string var from BASIC
	MOVE.L	#13,d2			;"TYPE MISMATCH" error preloaded
	CMP.B	#2,d0			;only string arguments allowed
	BNE	ErrorReport		;report error
	MOVE.L	a2,a3			;set up regs. for AssignString
	MOVE.L	a1,a2
	JSR	AssignString(a5)	;assign string desc. to string var
	BRA	CleanMem		;clean up memory and return to BASIC	
;##########################################################################


;##########################################################################
GiveInfo:
	LEA	InfoString,a0
	MOVE.L	#100,d0
	BSR	Info
	BSR	Pronounce	
	BRA	CleanMem
;##########################################################################

	
;LOCAL SUBROUTINES
;**************************************************************************

;----- Subroutines for parsing arguments ----------------------------------

GetIntegerArg:
	JSR	GetNextLibArg(a5) ;get argument from BASIC
	MOVE.L	#13,d2            ;"TYPE MISMATCH" error preloaded
	CMP.B	#3,d0             ;only numeric arguments allowed
	                          ; {d0 contains argument type}
	BLT	ErrorReport       ;report error
	JSR	IntegerArg(a5)    ;16 bit signed integer now in d3
	MOVE.L	#9,d2             ;"SUBSCRIPT OUT-OF-RANGE" error
	CMP.L	#0,d3             ;no negative inputs allowed
	BLT	ErrorReport       ;report error
	RTS
	
GetStringArg:
	JSR	GetNextLibArg(a5)   ;get argument from BASIC

ParseStringArg:
	MOVE.L	#13,d2              ;"TYPE MISMATCH" error preloaded
	CMP.B	#2,d0               ;only string arguments allowed
	                            ;  {d0 contains argument type}
	BNE	ErrorReport         ;report error
	MOVE.L	a2,a3               ;copy pointer into a3
	JSR	FreeTempDesc(a5)    ;free string space
	JSR	LoadStringDesc(a5)  ;on input a3 => string descriptor
	                            ;on output a3 => string data
	                            ;          d2 =  string length
RTS
;**************************************************************************

CloseItUp:

;------ Leave, but remain resident ----------------------------------------
;       ~~~~~~~~~~~~~~~~~~~~~~~~~~
	MOVE.L	#1,d0  ;exit with 0 in d0 makes code purgeable
	               ;          1 in d0 makes code resident
	RTS

;**************************************************************************

;-----Subroutine for generating a STR255 Exception Filename ---------------

GenerateFilename:
	MOVE.L	a2,a3			;input needs to be in a3
	JSR	LoadStringDesc(a5)	;move string descriptor into regs.
	MOVE.L	a3,a0			;set up regs for _BlockMove
	LEA	ExceptionFile,a1
	MOVE.B	d2,(a1)+		;push length byte
	MOVE.L	d2,d0

;PROCEDURE BlockMove (sourcePtr,destPtr:Ptr; byteCount:Size);	
	_BlockMove
	
	LEA	ExceptionFile,a0	;set up regs for _CmpString
	LEA	PhonemeExFile,a1
	MOVE.L	#$00080008,d0

;FUNCTION EqualString (aStr,bStr:Str255; caseSens,diacSens:BOOLEAN):BOOLEAN;	
	_CmpString

	BNE	ClearFlag		;if strings not equal then ClearFlag
	
SetFlag:
	LEA	PhonemeFlag,a1		;set flag = phoneme input
	ST	(a1)
	RTS		

ClearFlag:
	LEA	PhonemeFlag,a1		;clear flag = English input
	CLR.B	(a1)
	RTS

;**************************************************************************

;-----Subroutine for error handling ---------------------------------------

ErrorReport:
	BSR	QuitSpeech
	JSR	BasicError(a5)
	
;**************************************************************************
;**************************************************************************

;-----------STATIC DATA----------------------------------------------------
InfoString	DC.B	'DHIHS IHZ SPIY4CH LIH1B THRIY4 POYNT ZIY4ROW.'	
		DC.B	'RIH1TEHN BAY KEHVIH1N WAYLIY2 AHV AE2PUL DEY5TUN.#'
	.ALIGN	4
theSpeech	DC.L	0  	   ;storage for handle to speech globals
StringBuf	DC.L	0          ;buffer for string variable
phHandle	DC.L	0  	   ;storage for handle to phonetic codes
ExceptionFile	DCB.B	256,0	   ;storage for exception filename
PhonemeExFile	DC.B	8
		DC.B	'noReader' ;'noReader' filename		
PhonemeFlag	DC.B	0	   ;test flag 0:Phoneme input
                                   ;          =0:English input

END